import torch
from torch import nn

import os
import cv2
import json
import copy
import pickle
import random
import imutils
import operator
import numpy as np
import pandas as pd
from pathlib import Path
from itertools import chain
from functools import reduce
from functools import partial
from matplotlib import colors
from yaml import load, Loader
import matplotlib.pyplot as plt
from collections import OrderedDict
from configparser import ConfigParser
from imutils import resize as resize_img
from PIL import Image, ImageDraw, ImageFont
from typing import Iterable,Generator,Sequence,Iterator,List,Set,Dict,Union,Optional,Tuple



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/01_vision.ipynb.

# %% auto 0
__all__ = ['image_extensions', 'remove_images', 'bgr2rgb', 'rgb2bgr', 'gray2rgb', 'rgb2gray', 'rgb2rgba', 'bgra2rgb', 'rgba2rgb',
           'rgb_read', 'c1_read', 'tensor_to_img', 'plt_show', 'get_hw', 'get_image_files', 'to_pil', 'add_alpha',
           'has_alpha', 'color_to_rgb', 'solid_color_img', 'solid_color_img_like', 'get_pos_factors', 'get_pos_',
           'get_pos', 'paste_img']

# %% ../nbs/01_vision.ipynb 3
from .imports import *
from .core import *

# %% ../nbs/01_vision.ipynb 4
image_extensions = {'.art','.bmp','.cdr','.cdt','.cpt','.cr2','.crw','.djv','.djvu','.erf','.gif','.ico',
                    '.ief','.jng','.jp2','.jpe','.jpeg','.jpf','.jpg','.jpg2','.jpm','.jpx','.nef','.orf',
                    '.pat','.pbm','.pcx','.pgm','.png','.pnm','.ppm','.psd','.ras','.rgb','.svg','.svgz',
                    '.tif','.tiff','.wbmp','.xbm','.xpm','.xwd','.webp'}

def remove_images(images_path, final_count=0, fn=lambda x:True):
    imgs = get_image_files(images_path)
    remove = max(0, len(imgs) - final_count)
    removed = 0
    for i in imgs:
        if removed >= remove:
            break
        if fn(i):
            os.remove(i)
            removed+=1

def bgr2rgb(img):
    return cv2.cvtColor(img,cv2.COLOR_BGR2RGB)

def rgb2bgr(img):
    return cv2.cvtColor(img,cv2.COLOR_RGB2BGR)

def gray2rgb(img):
    return cv2.cvtColor(img,cv2.COLOR_GRAY2RGB)

def rgb2gray(img):
    return cv2.cvtColor(img,cv2.COLOR_RGB2GRAY)

def rgb2rgba(img):
    return cv2.cvtColor(img,cv2.COLOR_RGB2RGBA)

def bgra2rgb(img):
    if len(img.shape) > 2 and img.shape[2] == 4:
        return cv2.cvtColor(img,cv2.COLOR_BGRA2RGB)

def rgba2rgb(img):
    if len(img.shape) > 2 and img.shape[2] == 4:
        return cv2.cvtColor(img, cv2.COLOR_BGRA2BGR)

def rgb_read(img, shape=None):
    if not path_or_str(img):
        return img
    img = bgr2rgb(cv2.imread(str(img)))
    if shape is not None:
        img = cv2.resize(img, (shape[1], shape[0]))
    return img

def c1_read(img):
    "Read an image from `img` and convert it to grayscale."
    return cv2.imread(str(img), 0)

def tensor_to_img(t):
    if t.dim() > 3:
        return [np.array(np.transpose(t_,(1,2,0))) for t_ in t]
    return np.array(np.transpose(t,(1,2,0)))

def plt_show(im, cmap=None, title='', figsize=(7,7)):
    if path_or_str(im):
        im = rgb_read(im)
    if is_tensor(im):
        im = tensor_to_img(im)
        if is_list(im): im = im[0]
    fig=plt.figure(figsize=figsize)
    plt.imshow(im, cmap=cmap)
    plt.title(title)
    plt.show()

def get_hw(x):
    "Return the height and width of `x`."
    if is_clip(x):
        return x.size[1], x.size[0]
    return np.array(x).shape[:2]

def get_image_files(path, recurse=True, folders=None,
                    map_fn=None, # A function to apply to each item.
                    sort_key=None, # A function to sort the items by.
                    reverse=False, shuffle=False):
    "Get image files in `path` recursively, only in `folders`, if specified."
    l = get_files(path, extensions=image_extensions, recurse=recurse, folders=folders)
    if sort_key is not None:
        l = sorted(l, key=sort_key)
    if reverse:
        l = list(l)[::-1]
    if shuffle:
        random.shuffle(l)
    if map_fn is not None:
        return list_map(l, map_fn)
    return l

def to_pil(x):
    if is_array(x):
        return Image.fromarray(x)
    return x

def add_alpha(img, alpha=None):
    "Add an alpha channel to `img`."
    if alpha is None:
        alpha = 255
    img = to_pil(img)
    img.putalpha(alpha)
    return np.array(img)

def has_alpha(img):
    return np.array(img).shape[-1] == 4

def color_to_rgb(color):
    "Convert str `color` to RGB."
    if type(color) == str:
        return list_map(np.ceil(colors.to_rgb(color)).astype(int)*255, int)
    return list_map(color, int)

def solid_color_img(shape=(300,300,3), color='black', alpha=None):
    if len(shape) < 3:
        shape = tuple(shape) + (3,)
    image = np.zeros(shape, np.uint8)
    color = color_to_rgb(color)
    image[:] = color[:3]
    if len(color) == 4:
        alpha = color[3]
    if alpha is not None:
        image = Image.fromarray(image)
        image.putalpha(alpha)
        image = np.array(image)
    return image

def solid_color_img_like(img, color='black', alpha=None):
    h,w = np.array(img).shape[:2]
    shape = (h,w,3)
    return solid_color_img(shape=shape, color=color, alpha=alpha)

def get_pos_factors(pos):
    pos_dict = {'center':0.5, 'left':0.15, 'right':0.85, 'top':0.15, 'bottom':0.85}
    x_factor, y_factor = [pos_dict.get(p,p) for p in pos]
    return x_factor, y_factor

def get_pos_(bh, bw, h, w, x_factor, y_factor):
    x = int((bw - w) * x_factor)
    y = int((bh - h) * y_factor)
    return x,y

def get_pos(img, bg,
            pos=('center', 'center'), # int for exact position, float for relative position, or str for position in ['center', 'left', 'right', 'top', 'bottom']
            relative=True):
    "Get the position (x,y) of `img` on `bg` with `pos` as reference."
    p0,p1 = copy.deepcopy(pos)
    if is_int(p0) and is_int(p1):
        relative = False
    if not relative:
        return pos
    x_factor, y_factor = get_pos_factors(pos)
    bg = np.array(bg)
    img = np.array(img)
    bh,bw = bg.shape[:2]
    try:
        h,w = img.shape[:2]
    except:
        h,w = img[:2]
    x,y = get_pos_(bh, bw, h, w, x_factor, y_factor)
    if is_int(p0):
        x = p0
    if is_int(p1):
        y = p1
        # print(x, y)
    return x,y

def paste_img(img, bg, pos=('center', 'center'), relative=True):
    "Paste `img` on `bg` at `pos`."
    bg = np.array(copy.deepcopy(bg))
    img = np.array(copy.deepcopy(img))
    if bg.shape[-1] != 4:
        bg = add_alpha(bg)
    if img.shape[-1] != 4:
        img = add_alpha(img)
    h,w = img.shape[:2]
    bgh, bgw = bg.shape[:2]
    x,y = get_pos(img, bg, pos, relative=relative)
    
    r1,r2 = y,y+h
    c1,c2 = x,x+w
   
    if x < 0 or y < 0 or x >= bgw or y >= bgh:
        # print(x,y,h,w,bgh,bgw)
        bgh2 = max(bgh,h)
        bgw = max(bgw,w)
        px, py = 0.,0.
        if x < 0:
            bgw2 = bgw + abs(x)
            x = 0
            px = 1.
        elif x >= bgw:
            bgw2 = x+w
        else:
            bgw2 = bgw
        if y < 0:
            bgh2 = bgh + abs(y)
            y = 0
            py = 1.
        elif y >= bgh:
            bgh2 = y+h
        else:
            bgh2 = bgh
        new_bg = solid_color_img((bgh2, bgw2, 3), alpha=0)
        bg = paste_img(bg, new_bg, pos=(px,py))
        bgh, bgw = bg.shape[:2]
        r1,r2 = y,y+h
        c1,c2 = x,x+w
    
    elif h>bgh and w>bgw:
        hf = (h-bgh)//2
        wf = (w-bgw)//2
        img = img[hf:hf+bgh, wf:wf+bgw]
        r1,r2 = 0,bgh
        c1,c2 = 0,bgw
    elif h>bgh:
        hf = (h-bgh)//2
        img = img[hf:hf+bgh]
        r1,r2 = 0,bgh
    elif w>bgw:
        wf = (w-bgw)//2
        img = img[:,wf:wf+bgw]
        c1,c2 = 0,bgw

    bg2 = copy.deepcopy(bg[r1:r2, c1:c2])
    bg2h, bg2w = get_hw(bg2)
    img = img[:bg2h, :bg2w]
    # print(bg.shape, bg2.shape, img.shape, r2-r1, c2-c1)
    bgh, bgw = bg2.shape[:2]
    bg_rgb = bg2[...,:3]
    img_rgb = img[...,:3]
    bg_alpha = bg2[...,3]/255.0
    img_alpha = img[...,3]/255.0
    # print(img_alpha.shape)
    if (bg_alpha == 1).all():
        bg_alpha = 1-img_alpha
    i_len = len(img_alpha)
    for i in range(i_len):
        j_len = len(img_alpha[i])
        for j in range(j_len):
            if img_alpha[i][j] == 1:
                # try:
                bg_alpha[i][j] = 0
                # except:
                    # pass
    out_alpha = bg_alpha + img_alpha*(1-bg_alpha)
    out_rgb = (bg_rgb*bg_alpha[...,np.newaxis] + img_rgb*img_alpha[...,np.newaxis]*(1-bg_alpha[...,np.newaxis])) / out_alpha[...,np.newaxis]
    out_rgba = np.dstack((out_rgb,out_alpha*255)).astype(np.uint8)
    bg[r1:r2,c1:c2] = out_rgba
    return bg



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/00_core.ipynb.

# %% auto 0
__all__ = ['default_device', 'flatten_list', 'noop', 'is_list', 'is_tuple', 'list_or_tuple', 'is_iter', 'is_dict', 'is_df',
           'is_str', 'is_int', 'is_float', 'is_array', 'is_pilimage', 'is_tensor', 'is_set', 'is_path', 'path_or_str',
           'is_norm', 'params', 'is_frozen', 'is_unfrozen', 'is_subscriptable', 'is_sequential', 'is_clip', 'path_name',
           'path_stem', 'path_suffix', 'extend_path_name', 'end_of_path', 'add_ext_to_path', 'last_modified',
           'load_yaml', 'save_obj', 'load_obj', 'yml_to_pip', 'set_pip_req', 'merge_dicts', 'dict_values', 'dict_keys',
           'sort_dict', 'locals_to_params', 'list_map', 'next_batch', 'model_children', 'replace_dict_key', 'proc_fn',
           'filter_dict', 'setify', 'get_files']

# %% ../nbs/00_core.ipynb 3
from .imports import *

# %% ../nbs/00_core.ipynb 4
def default_device(device=None):
    if device is None:
        device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
    return device

def flatten_list(l):
    "Flatten a list of lists."
    l2 = []
    for x in l:
        if is_list(x):
            l2 += flatten_list(x)
        else:
            l2.append(x)
    return l2

def noop(x=None, **kwargs):
    "Do nothing."
    return x

def is_list(x):
    return isinstance(x, list)

def is_tuple(x):
    return isinstance(x, tuple)

def list_or_tuple(x):
    return (is_list(x) or is_tuple(x))

def is_iter(o):
    "Test whether `o` can be used in a `for` loop."
    #Rank 0 tensors in PyTorch are not really iterable
    return isinstance(o, (Iterable,Generator)) and getattr(o,'ndim',1)

def is_dict(x):
    return isinstance(x, dict)

def is_df(x):
    return isinstance(x, pd.core.frame.DataFrame)

def is_str(x):
    return isinstance(x, str)

def is_int(x):
    return isinstance(x, int)    

def is_float(x):
    return isinstance(x, float)

def is_array(x):
    return isinstance(x, np.ndarray)

def is_pilimage(x):
    return 'PIL' in str(type(x))

def is_tensor(x):
    return isinstance(x, torch.Tensor)

def is_set(x):
    return isinstance(x, set)

def is_path(x):
    return isinstance(x, Path)

def path_or_str(x):
    return is_str(x) or is_path(x)

def is_norm(x):
    return type(x).__name__ == 'Normalize'

def params(m):
    "Return all parameters of `m`."
    return [p for p in m.parameters()]

def is_frozen(model):
    return np.array([not p.requires_grad for p in (params(model))]).all()

def is_unfrozen(model):
    return np.array([p.requires_grad for p in (params(model))]).all()

def is_subscriptable(x):
    return hasattr(x, '__getitem__')

def is_sequential(x):
    return isinstance(x, nn.Sequential)

def is_clip(x):
    return type(x).__name__ == 'ProntoClip' or 'moviepy' in str(type(x))

def path_name(x):
    return Path(x).name

def path_stem(x):
    return Path(x).stem

def path_suffix(path):
    return Path(path).suffix

def extend_path_name(p, s='_2'):
    "Add `s` to the name of a path `p`. Before the extension."
    p = Path(p)
    return p.parent/(p.stem+s+p.suffix)

def end_of_path(p, n=2):
    "Get the last `n` parts of a path `p`."
    parts = p.parts
    p = Path(parts[-n])
    for i in range(-(n-1), 0):
        p/=parts[i]
    return p

def add_ext_to_path(p, ext='pkl'):
    "Add an extension to a path `p` if it doesn't have one."
    if ext[0] != '.':
        ext = '.'+ext
    if len(Path(p).suffix) == 0:
        p = str(p)
        if p[-1] != '.':
            p+=ext
        else:
            p+=ext[1:]
    return p

def last_modified(x):
    "Get the last modified time of a file."
    return x.stat().st_ctime

def load_yaml(file):
    with open(file) as f:
        env = load(f, Loader=Loader)
    return env

def save_obj(path, obj):
    with open(path, 'wb') as f:
        pickle.dump(obj, f, pickle.HIGHEST_PROTOCOL)

def load_obj(path):
    with open(path, 'rb') as f:
        return pickle.load(f)
    
def yml_to_pip(yml, remove_eq=True):
    "Get pip packages from a conda environment `yml` file."
    env = load_yaml(yml)
    env_pip = env['dependencies'][-1]['pip']
    pip_list = []
    for x in env_pip:
        if remove_eq:
            x = x.split('==')[0].split('>=')[0]
        if 'nvidia' not in x:
            pip_list.append(x)
    # if remove_eq:
        # env_pip = [x.split('==')[0].split('>=')[0] for x in env_pip]
    return " ".join(env_pip)

def set_pip_req(yml, settings, remove_eq=True):
    "Update the pip_requirements in settings.ini from a conda environment `yml` file."
    env_pip = yml_to_pip(yml, remove_eq=remove_eq)
    config = ConfigParser(delimiters=['='], allow_no_value=True)
    config.read(settings)
    cfg = config['DEFAULT']
    config.set('DEFAULT', 'pip_requirements', env_pip)
    with open(settings, 'w') as configfile:
        config.write(configfile)

def merge_dicts(d1,d2):
    d = {}
    for k in d1:
        d[k] = d1[k]
    for k in d2:
        d[k] = d2[k]
    return d

def dict_values(d):
    "Get the values of a dictionary sorted by key."
    return [v for _,v in sorted(d.items(), key=lambda x:x[0])]

def dict_keys(d):
    "Get the sorted keys of a dictionary."
    return [k for k,_ in sorted(d.items(), key=lambda x:x[0])]

def sort_dict(d, by_value=False):
    "Sort a dictionary by key by default or by value if `by_value` is True."
    idx = int(by_value)
    return {k: v for k, v in sorted(d.items(), key=lambda item:item[idx])}

def locals_to_params(l, omit=[], expand=['kwargs']):
    "Convert all the local variables to a dictionary of parameters."
    if 'kwargs' not in expand:
        expand.append('kwargs')
    l = copy.deepcopy(l)
    if 'self' in l.keys():
        del l['self']
    if '__class__' in l.keys():
        del l['__class__']
    keys = dict_keys(l)
    for k in keys:
        if k in expand:
            for k2 in l[k]:
                if k2 not in l.keys():
                    l[k2] = l[k][k2]
            del l[k]
        if k in omit:
            del l[k]
    return l

def list_map(l, m):
    "Apply `m` to each element of `l`."
    return list(pd.Series(l).apply(m))

def next_batch(dl):
    "Get the next batch from a dataloader `dl`."
    return next(iter(dl))

def model_children(model):
    "Get the children of a model."
    return list(model.children())

def replace_dict_key(d:dict, x='', y='',
                     strict=False): # If True, replace if `x` == key. If False, replace if `x` in key.
    "Replace key `x` with `y` in dictionary `d`."
    if (x == '' and y == '') or x == y:
        return d
    if strict:
        fn = lambda k: k.replace(x, y) if x == k else k
    else:
        fn = lambda k: k.replace(x, y) if x in k else k
    return {fn(k): v for k, v in d.items()}

def proc_fn(fn):
    "Process function `fn`. It can be a string to match or a function that takes a key and returns True/False."
    if is_str(fn):
        t = copy.deepcopy(fn)
        fn = lambda x: x==t
    return fn

def filter_dict(d,
                fn): # Can be a string to match or a function that takes a key and returns True/False.
    "Filter dict `d` based on function `fn`."
    d2 = {}
    fn = proc_fn(fn)
    keys = dict_keys(d)
    for k in keys:
        if fn(k):
            d2[k] = d[k]
    return d2

def setify(o): return o if isinstance(o,set) else set(list(o))

def _get_files(p, fs, extensions=None):
    p = Path(p)
    res = [p/f for f in fs if not f.startswith('.')
           and ((not extensions) or f'.{f.split(".")[-1].lower()}' in extensions)]
    return res

def get_files(path, extensions=None, recurse=True, folders=None, followlinks=True, make_str=False):
    "Get all the files in `path` with optional `extensions`, optionally with `recurse`, only in `folders`, if specified."
    if folders is None:
        folders = list([])
    path = Path(path)
    if extensions is not None:
        extensions = setify(extensions)
        extensions = {e.lower() for e in extensions}
    if recurse:
        res = []
        for i,(p,d,f) in enumerate(os.walk(path, followlinks=followlinks)): # returns (dirpath, dirnames, filenames)
            if len(folders) !=0 and i==0: d[:] = [o for o in d if o in folders]
            else:                         d[:] = [o for o in d if not o.startswith('.')]
            if len(folders) !=0 and i==0 and '.' not in folders: continue
            res += _get_files(p, f, extensions)
    else:
        f = [o.name for o in os.scandir(path) if o.is_file()]
        res = _get_files(path, f, extensions)
    if make_str: res = [str(o) for o in res]
    return list(res)
